@@ -1,5 +1,5 @@
 /*
- * Copyright 2013 University Corporation for Atmospheric Research
+ * Copyright 2020 University Corporation for Atmospheric Research
  *
  * This file is part of the UDUNITS-2 package.  See the file COPYRIGHT
  * in the top-level source-directory of the package for copying and
@@ -14,9 +14,7 @@
 
 #include <config.h>
 
-#ifndef	_XOPEN_SOURCE
-#   define _XOPEN_SOURCE 500
-#endif
+#include "udunits2.h"
 
 #if defined(__linux__) || defined(__CYGWIN__)
 #   ifndef _GNU_SOURCE
@@ -48,15 +46,14 @@
 #elif defined _WIN32
 #include <windows.h>
 #endif
+#include <expat.h>
 
 #ifndef DLL_UDUNITS2
-#define XML_STATIC
+#   define XML_STATIC
 #endif
-#include "expat.h"
-#include "udunits2.h"
 
-#ifndef _XOPEN_PATH_MAX
-#   define _XOPEN_PATH_MAX 1024 // Includes terminating NUL
+#ifndef PATH_MAX
+#   define PATH_MAX 4096 // Includes terminating NUL
 #endif
 
 #define NAME_SIZE 128
@@ -581,7 +578,7 @@ mapUnitToId(
  *
  * Arguments:
  *      unit            Pointer to the unit.
- *      id              Pointer to the identifier upon wich to base all
+ *      id              Pointer to the identifier upon which to base all
  *                      derived identifiers.
  *      encoding        The encoding of "id".
  *      isName          Whether or not "id" is a name.
@@ -624,7 +621,7 @@ mapUnitToIds(
  *
  * Arguments:
  *      unit            Pointer to the unit.
- *      name            Pointer to the name upon wich to base all derived names.
+ *      name            Pointer to the name upon which to base all derived names.
  *      encoding        The encoding of "name".
  * Returns:
  *      0               Failure.
@@ -645,7 +642,7 @@ mapUnitToNames(
  *
  * Arguments:
  *      unit            Pointer to the unit.
- *      symbol          Pointer to the symbol upon wich to base all derived
+ *      symbol          Pointer to the symbol upon which to base all derived
  *                      symbols.
  *      encoding        The encoding of "symbol".
  * Returns:
@@ -1807,7 +1804,7 @@ static void
 endImport(
     void*		data)
 {
-    char        buf[_XOPEN_PATH_MAX];
+    char        buf[PATH_MAX];
     const char* path;
 
     if (text[0] == '/') {
@@ -2090,7 +2087,7 @@ readXml(
         ut_handle_error_message("Couldn't create XML parser");
     }
     else {
-        char base[_XOPEN_PATH_MAX];
+        char base[PATH_MAX];
 #ifdef _MSC_VER
         {
             char drive[_MAX_DRIVE+1]; // Will have trailing colon
@@ -2134,40 +2131,64 @@ readXml(
 static const char*
 default_udunits2_xml_path()
 {
-    char const * soname = 0;
-#if defined(__APPLE__) || defined(__linux__) || defined(__CYGWIN__)
-    #define END_BIT "/share/udunits/udunits2.xml"
-    #define SEP '/'
-    Dl_info info;
-    if (dladdr(default_udunits2_xml_path, &info)) {
-        soname = info.dli_fname;
-    }
-#elif defined(_WIN32)
-    #define END_BIT "\\share\\udunits\\udunits2.xml"
-    #define SEP '\\'
-    // NB: XP+ solution!
-    HMODULE hModule = NULL;
-    GetModuleHandleEx(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS,
-            (LPCTSTR)default_udunits2_xml_path, &hModule);
-    char tmpbuf[MAX_PATH * 4];
-    GetModuleFileName(hModule, &tmpbuf[0], MAX_PATH * 4);
-    soname = &tmpbuf[0];
-#endif
-    if (soname) {
-        char * result = malloc(strlen(soname) + strlen(END_BIT) + 1);
-        if (result) {
-            strcpy(result, soname);
-            if (strrchr(result, SEP)) {
-                *strrchr(result, SEP) = '\0';
-                if (strrchr(result, SEP)) {
-                    *strrchr(result, SEP) = '\0';
-                }
+    // Returned absolute pathname of XML database
+    static char absXmlPathname[PATH_MAX];
+
+    if (absXmlPathname[0] == 0) {
+        const char* prefix = NULL; // Installation directory
+
+#       if defined(__APPLE__) || defined(__linux__) || defined(__CYGWIN__)
+            Dl_info     info;
+            const char  sep = '/'; // Pathname component separator
+            char        buf[PATH_MAX];
+            const char  relXmlPathname[] = "share/udunits/udunits2.xml";
+
+            // The following should get pathname of shared-library
+            if (!dladdr(default_udunits2_xml_path, &info)) {
+                prefix = NULL;
             }
-            strcat(result, END_BIT);
-            return result;
-        }
-    }
-    return DEFAULT_UDUNITS2_XML_PATH;
+            else {
+                strncpy(buf, info.dli_fname, sizeof(buf))[sizeof(buf)-1] = 0;
+                memmove(buf, dirname(buf), sizeof(buf)); // "lib"
+                memmove(buf, dirname(buf), sizeof(buf)); // "lib/.."
+                prefix = buf;
+            }
+#       elif defined(_WIN32)
+            const char sep = '\\'; // Pathname component separator
+            char       buf[MAX_PATH * 4];
+            const char relXmlPathname[] = "share\\udunits\\udunits2.xml";
+            HMODULE    hModule = NULL;
+
+            GetModuleHandleEx(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS,
+                    (LPCTSTR)default_udunits2_xml_path, &hModule);
+
+            GetModuleFileName(hModule, buf, sizeof(buf)); // Library pathname
+            *strrchr(buf, sep) = 0; // "lib"
+            *strrchr(buf, sep) = 0; // "lib/.."
+
+            prefix = buf;
+#       endif
+
+        if (prefix == NULL) {
+            strncpy(absXmlPathname, DEFAULT_UDUNITS2_XML_PATH,
+                    sizeof(absXmlPathname));
+        } // Use default pathname
+        else {
+            int prefixLen = strlen(prefix);
+            if (prefix[prefixLen-1] == sep) {
+                --prefixLen;
+                if (prefix[prefixLen-1] == sep)
+                    --prefixLen;
+            }
+
+            snprintf(absXmlPathname, sizeof(absXmlPathname), "%.*s%c%s",
+                    prefixLen, prefix, sep, relXmlPathname);
+        } // Use computed pathname
+
+        absXmlPathname[sizeof(absXmlPathname)-1] = 0; // Ensure NUL terminated
+    } // `absXmlPathname` not set
+
+    return absXmlPathname;
 }
 
 /**
@@ -2179,7 +2200,7 @@ default_udunits2_xml_path()
  *                  pathname specified by the environment variable
  *                  UDUNITS2_XML_PATH is returned if set; otherwise, the
  *                  compile-time pathname of the installed, default, unit
- *                  database is returned.
+ *                  database is returned. Caller must not free.
  */
 const char*
 ut_get_path_xml(
